/*!
\file
\verbatim

Copyright (c) 2004, Sylvain Paris and Francois Sillion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    * Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    * Neither the name of ARTIS, GRAVIR-IMAG nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

\endverbatim
 *  These files contain code made by Sylvain Paris under supervision of
 * Franois Sillion for his PhD work with <a
 * href="http://artis.imag.fr">ARTIS project</a>. ARTIS is a
 * research project in the GRAVIR/IMAG laboratory, a joint unit of
 * CNRS, INPG, INRIA and UJF.
 *
 *  Source code can be downloaded from
 * <a href="http://artis.imag.fr/Members/Sylvain.Paris/index.html#code">Sylvain Paris' page</a>. <!--'-->
 *
 *  This file provides various "traits" to be used with image_file.h.
 */


#ifndef __IMAGE_CONTAINER_TRAITS__
#define __IMAGE_CONTAINER_TRAITS__

#include <algorithm>

#include<qimage.h>
#include<qpixmap.h>

#include "array.h"
#include "channel_image.h"
#include "color.h"

#include "image_file.h"

namespace Image_file{

  /*

    ##############
    # QT version #
    ##############
  
  */


  template<>
  void load(const char*    file_name,
	    QPixmap* const target){
    
    target->load(QString(file_name));
  }


  
  template<>
  void save(const char*    file_name,
	    const QPixmap& source){

    const std::string format = guess_format(file_name);

    if (format == std::string()) {
      char* escape_format = strdup(QImage::outputFormats().first());

      Message::warning<<"output format not recognized. "
		      <<"image written ("<<file_name<<") in format : "
		      <<escape_format<<Message::done;

      source.save(QString(file_name),escape_format);

      free(escape_format);

    }
    else{ 
      source.save(QString(file_name),format.c_str());
    }    
  }
  
  template<>
  void load(const char*   file_name,
	    QImage* const target){
    
    target->load(QString(file_name));
  }


  
  template<>
  void save(const char*   file_name,
	    const QImage& source){

    const std::string format = guess_format(file_name);

    if (format == std::string()) {
      char* escape_format = strdup(QImage::outputFormats().first());

      Message::warning<<"output format not recognized. "
		      <<"image written ("<<file_name<<") in format : "
		      <<escape_format<<Message::done;

      source.save(QString(file_name),escape_format);

      free(escape_format);

    }
    else{
      source.save(QString(file_name),format.c_str());
    }    
  }

  
  
  namespace Image_container_traits{

    /*

      #############################
      # RGB_channel_image version #
      #############################
    
    */

    template<typename Real>
    inline unsigned int
    get_width(const RGB_channel_image<Real>& container){
      return container.x_size();
    }

    
    template<typename Real>
    inline unsigned int
    get_height(const RGB_channel_image<Real>& container){
      return container.y_size();
    }

    
    template<typename Real>
    inline void
    set_size(RGB_channel_image<Real>* const container,
	     const unsigned int        width,
	     const unsigned int        height){
      container->resize(width,height);
    }

    
    template<typename Real>
    inline void
    set_value(RGB_channel_image<Real>* const container,
	      const unsigned int       x,
	      const unsigned int       y,
	      const double             r,
	      const double             g,
	      const double             b){
    
      (*container)[RGB_channel_image<Real>::red](x,y)   = r;
      (*container)[RGB_channel_image<Real>::green](x,y) = g;
      (*container)[RGB_channel_image<Real>::blue](x,y)  = b;
    }

  
    template<typename Real>
    inline void
    get_value(const RGB_channel_image<Real>& container,
	      const unsigned int             x,
	      const unsigned int             y,
	      double* const                  r,
	      double* const                  g,
	      double* const                  b){

      *r = container[RGB_channel_image<Real>::red](x,y);
      *g = container[RGB_channel_image<Real>::green](x,y);
      *b = container[RGB_channel_image<Real>::blue](x,y);
    }



    
   /*

      #############################
      # Array_2D<Color,A> version #
      #############################
    
    */

    template<typename Color,typename A>
    inline unsigned int
    get_width(const Array_2D<Color,A>& container){
      return container.x_size();
    }

    
    template<typename Color,typename A>
    inline unsigned int
    get_height(const Array_2D<Color,A>& container){
      return container.y_size();
    }

    
    template<typename Color,typename A>
    inline void
    set_size(Array_2D<Color,A>* const container,
	    const unsigned int        width,
	    const unsigned int        height){
      container->resize(width,height);
    }

    
    template<typename Color,typename A>
    inline void
    set_value(Array_2D<Color,A>* const container,
	      const unsigned int       x,
	      const unsigned int       y,
	      const double             r,
	      const double             g,
	      const double             b){
    
      (*container)(x,y) = Color(r,g,b);
    }

  
    template<typename Color,typename A>
    inline void
    get_value(const Array_2D<Color,A>&    container,
	      const unsigned int          x,
	      const unsigned int          y,
	      double* const               r,
	      double* const               g,
	      double* const               b){

      container(x,y).get_RGB(r,g,b);
    }


  
    /*

      ##################
      # double version #
      ##################
    
    */


    template<typename A>
    inline unsigned int
    get_width(const Array_2D<double,A>& container){
      return container.x_size();
    }

    
    template<typename A>
    inline unsigned int
    get_height(const Array_2D<double,A>& container){
      return container.y_size();
    }

    
    template<typename A>
    inline void
    set_size(Array_2D<double,A>* const container,
	    const unsigned int        width,
	    const unsigned int        height){
      container->resize(width,height);
    }

    
    template<typename A>
    inline void
    set_value(Array_2D<double,A>* const container,
	      const unsigned int        x,
	      const unsigned int        y,
	      const double              r,
	      const double              g,
	      const double              b){
    
      (*container)(x,y) = std::max(std::max(r,g),b);
    }

  
    template<typename A>
    inline void
    get_value(const Array_2D<double,A>& container,
	      const unsigned int        x,
	      const unsigned int        y,
	      double* const             r,
	      double* const             g,
	      double* const             b){

      *r = container(x,y);
      *g = *r;
      *b = *r;
    }


    /*

      #################
      # float version #
      #################
    
    */


    template<typename A>
    inline unsigned int
    get_width(const Array_2D<float,A>& container){
      return container.x_size();
    }

    
    template<typename A>
    inline unsigned int
    get_height(const Array_2D<float,A>& container){
      return container.y_size();
    }

    
    template<typename A>
    inline void
    set_size(Array_2D<float,A>* const container,
	    const unsigned int        width,
	    const unsigned int        height){
      container->resize(width,height);
    }

  
    template<typename A>
    inline void
    set_value(Array_2D<float,A>* const container,
	      const unsigned int       x,
	      const unsigned int       y,
	      const double             r,
	      const double             g,
	      const double             b){
    
      (*container)(x,y) = static_cast<float>(std::max(std::max(r,g),b));
    }

  
    template<typename A>
    inline void
    get_value(const Array_2D<float,A>&    container,
	      const unsigned int          x,
	      const unsigned int          y,
	      double* const               r,
	      double* const               g,
	      double* const               b){

      *r = static_cast<double>(container(x,y));
      *g = *r;
      *b = *r;
    }


    /*

      ################
      # char version #
      ################
    
    */


    template<typename A>
    inline unsigned int
    get_width(const Array_2D<char,A>& container){
      return container.x_size();
    }

    
    template<typename A>
    inline unsigned int
    get_height(const Array_2D<char,A>& container){
      return container.y_size();
    }

    
    template<typename A>
    inline void
    set_size(Array_2D<char,A>* const container,
	    const unsigned int        width,
	    const unsigned int        height){
      container->resize(width,height);
    }
    
    template<typename A>
    inline void
    set_value(Array_2D<char,A>* const container,
	      const unsigned int      x,
	      const unsigned int      y,
	      const double            r,
	      const double            g,
	      const double            b){
    
      (*container)(x,y) = static_cast<char>(255*std::max(std::max(r,g),b));
    }

  
    template<typename A>
    inline void
    get_value(const Array_2D<char,A>&    container,
	      const unsigned int         x,
	      const unsigned int         y,
	      double* const              r,
	      double* const              g,
	      double* const              b){

      *r = static_cast<double>(container(x,y)) / 255.0;
      *g = *r;
      *b = *r;
    }


    /*

      ################
      # bool version #
      ################
    
    */


    template<typename A>
    inline unsigned int
    get_width(const Array_2D<bool,A>& container){
      return container.x_size();
    }

    
    template<typename A>
    inline unsigned int
    get_height(const Array_2D<bool,A>& container){
      return container.y_size();
    }

    
    template<typename A>
    inline void
    set_size(Array_2D<bool,A>* const container,
	    const unsigned int        width,
	    const unsigned int        height){
      container->resize(width,height);
    }

    
    template<typename A>
    inline void
    set_value(Array_2D<bool,A>* const container,
	      const unsigned int      x,
	      const unsigned int      y,
	      const double            r,
	      const double            g,
	      const double            b){
    
      (*container)(x,y) = ((r!=0.0)||(g!=0.0)||(b!=0.0));
    }

  
    template<typename A>
    inline void
    get_value(const Array_2D<bool,A>& container,
	      const unsigned int      x,
	      const unsigned int      y,
	      double* const           r,
	      double* const           g,
	      double* const           b){

      *r = container(x,y) ? 1.0 : 0.0;
      *g = *r;
      *b = *r;
    }

  }// END OF namespace  
}// END OF namespace

#endif
