/*! \file
\verbatim

Copyright (c) 2004, Sylvain Paris and Francois Sillion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    * Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    * Neither the name of ARTIS, GRAVIR-IMAG nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

\endverbatim

  These files contain code made by Sylvain Paris under supervision of
 Franois Sillion for his PhD work with <a
 href="http://artis.imag.fr">ARTIS project</a>. ARTIS is a
 research project in the GRAVIR/IMAG laboratory, a joint unit of
 CNRS, INPG, INRIA and UJF.
 
  Source code can be downloaded from
 <a href="http://artis.imag.fr/Members/Sylvain.Paris/index.html#code">Sylvain Paris' page</a>. <!--'-->

  This header contains the load() and save() functions that allow a
  simple manipulation of image file. These functions are based on the
  <a href="http://http://www.trolltech.com/products/qt/">QT
  library</a>.
  

  Image point (0,0) is located in the lower left corner.

  The load() and save() functions work with any 2D data structures
  containing "colors". To do so, you have to implement the functions
  set_value(), set_size(), get_value(), get_width(), get_height() as
  in the following example.


  <hr />

  To use the structure \e My_image that has a method \e pixel(x,y) to
  access the point (x,y) that provides the \e red, \e green and \e
  blue values. \e My_image also provides the variables \e width and \e
  height for its size. This size can be changed through the function
  \e resize(x,y).
  
  To use this image module, you have to write this file \e my_image_traits.h 

  \code
  // file : my_image_traits.h



  namespace Image_file{
    namespace Image_container_traits{

      void
      get_value(const My_image&    image,
	        const unsigned int x,
	        const unsigned int y,
	        double* const      r,
	        double* const      g,
	        double* const      b){

	// Fill *r, *g et *b from the data structure My_image.
		
        *r = image.pixel(x,y).red;
        *g = image.pixel(x,y).green;
        *b = image.pixel(x,y).blue;
      }

      
      unsigned int
      get_width(const My_image& image){

        // Return the image width.
	
        return width;
      }


      unsigned int
      get_height(const My_image& image){

        // Return the image height.
      
        return height;
      }
      
      
      void
      set_value(My_image* const    image,
	        const unsigned int x,
	        const unsigned int y,
	        const double       r,
	        const double       g,
	        const double       b){

        // Fill the data structure My_image with r, g et b.
	
        image->pixel(x,y).red   = r;
        image->pixel(x,y).green = g;
        image->pixel(x,y).blue  = b;
      }

      
      void
      set_size(My_image* const    image,
               const unsigned int width,
	       const unsigned int height){

	// Set the image.
	
	image->resize(width,height);       
      }
      
    }
  }
  \endcode

  Then you can easily used the following functions:
  \code
    #include "my_image_traits.h"
    #include "image_file.h"
    
    // ...
    
    My_image an_image;
    Image_file::load("a_file.jpeg",&an_image);
    Image_file::save("another_file.bmp",an_image);

    // ...
  \endcode
*/  

#ifndef __IMAGE_FILE__
#define __IMAGE_FILE__

#include <algorithm>
#include <functional>
#include <stdexcept>

#include <qcolor.h> // Must be linked with QT with correct include paths.
#include <qimage.h>
#include <qstring.h>

#include "msg_stream.h"
#include "tools.h"


namespace Image_file{
  
  //! Read an image file.
  template<typename Image_container>
  void load(const char* file_name,
	    Image_container* const target);
  
  //! Write an image file.
  template<typename Image_container>
  void save(const char* file_name,
	    const Image_container& source);


  //! Convert an image through supported data structures.
  template<typename Source_container,typename Target_container>
  inline void convert(const Source_container& source,
		      Target_container* const target);

  //! Guess the image format corresponding to a file name.
  inline std::string guess_format(const std::string filename);
  
  
  //! Flip the image horizontally.
  template<typename Image_container>
  inline void reverse_x(Image_container* target);


  //! Flip the image vertically.
  template<typename Image_container>
  inline void reverse_y(Image_container* target);


    
/*
  
  #############################################
  #############################################
  #############################################
  ######                                 ######
  ######   I M P L E M E N T A T I O N   ######
  ######                                 ######
  #############################################
  #############################################
  #############################################
  
*/


  namespace Image_container_traits{}


  
  template<typename Image_container>
  void load(const char*      file_name,
	    Image_container* const target){
    
    QImage qt_image = QImage(QString(file_name));

    if (qt_image.isNull()){
      
      Message::warning<<"failure while loading \""<<file_name<<"\""
		      <<Message::done;
      
      throw std::invalid_argument("error during image loading");
    }
    
    // A few typedefs to ease the following code.
    typedef unsigned int size_type;

    const size_type width  = qt_image.width();
    const size_type height = qt_image.height();    

    Image_container_traits::set_size(target,width,height);

    for(size_type x=0;x<width;x++){
      for(size_type y=0;y<height;y++){
	
	QRgb rgb       = qt_image.pixel(x,height-1-y);
	
	Image_container_traits::
	  set_value(target,x,y,
		    static_cast<double>(qRed(rgb))   / 255.0,
		    static_cast<double>(qGreen(rgb)) / 255.0,
		    static_cast<double>(qBlue(rgb))  / 255.0); 
      }
    }
  }


  /*! Try to guess the file format from the file name.  If the format
    cannot be determined, the first format in the QT list is used by
    default.  */ 
  template<typename Image_container>
  void save(const char* file_name,
	    const Image_container& source){
   
    typedef unsigned int size_type;

    const size_type width  = Image_container_traits::get_width(source);
    const size_type height = Image_container_traits::get_height(source);    

    // create a QImage
    QImage image(width,height,32);

    for(size_type x=0;x<width;x++){
      for(size_type y=0;y<height;y++){
	double r,g,b;
	
	Image_container_traits::get_value(source,x,height-1-y,&r,&g,&b);
	
	image.setPixel(x,y,qRgb(static_cast<char>(r*255),
				static_cast<char>(g*255),
				static_cast<char>(b*255)));
      }
    }

    // Try to guess the file format.
    const std::string format = guess_format(file_name);

    if (format == std::string()) {
      // If format not found, use the default and save however.
      char* escape_format = strdup(QImage::outputFormats().first());

      Message::warning<<"output format not recognized. "
		      <<"image written ("<<file_name<<") in format: "
		      <<escape_format<<Message::done;

      image.save(QString(file_name),escape_format);

      free(escape_format);

    }
    else{ // If format is guessed.
      image.save(QString(file_name),format.c_str());
    }
  }



  
  template<typename Source_container,typename Target_container>
  void convert(const Source_container& source,
	       Target_container* const target){

    typedef unsigned int size_type;

    const size_type width  = Image_container_traits::get_width(source);
    const size_type height = Image_container_traits::get_height(source);    

    Image_container_traits::set_size(target,width,height);

    for(size_type x=0;x<width;x++){
      for(size_type y=0;y<height;y++){
	
	double r,g,b;
	
	Image_container_traits::get_value(source,x,y,&r,&g,&b);
	Image_container_traits::set_value(target,x,y,r,g,b);
      }
    }

  }


  
  std::string guess_format(const std::string filename){

    using namespace std;
    
    // Conversion into a STL data stricture.
    list<string> valid_answers;
    QStrList qlist = QImage::outputFormats();
    for(char* c=qlist.first();c!=NULL;c=qlist.next()){
      valid_answers.push_back(c);
    }
    
    // Add some more if needed.
    const bool has_JPG =
      (find(valid_answers.begin(),valid_answers.end(),"JPG")
       !=valid_answers.end());
    
    const bool has_JPEG =
      (find(valid_answers.begin(),valid_answers.end(),"JPEG")
       !=valid_answers.end());
    
    if (has_JPG&&(!has_JPEG)){
      valid_answers.push_back("JPEG");
    }
    
    if (has_JPEG&&(!has_JPG)){
      valid_answers.push_back("JPG");
    }

    // Look for the extensions.
    const string query = Tools::to_upper_case(filename);
    const unsigned int query_length = query.length();
    for(list<string>::iterator f=valid_answers.begin();
	f!=valid_answers.end();f++){

      if (query.find(*f)<query_length){
	if ((*f=="JPG")&&(!has_JPG)){
	  return "JPEG";
	}
	if ((*f=="JPEG")&&(!has_JPEG)){
	  return "JPG";
	}
	return *f;
      }
    }
    
    return string();

  }


  template<typename Image_container>
  void reverse_x(Image_container* target){
    
    typedef unsigned int size_type;

    const size_type width  = Image_container_traits::get_width(*target);
    const size_type height = Image_container_traits::get_height(*target);    

    for(size_type x=0;x<(width+1)/2;x++){
      for(size_type y=0;y<height;y++){
	double r1,g1,b1,r2,g2,b2;
	Image_container_traits::get_value(*target,width-1-x,y,&r1,&g1,&b1);
	Image_container_traits::get_value(*target,x,y,&r2,&g2,&b2);
	Image_container_traits::set_value(target,x,y,r1,g1,b1);
	Image_container_traits::set_value(target,width-1-x,y,r2,g2,b2);
      }
    }
  }

  
  template<typename Image_container>
  void reverse_y(Image_container* target){
    
    typedef unsigned int size_type;

    const size_type width  = Image_container_traits::get_width(*target);
    const size_type height = Image_container_traits::get_height(*target);    

    for(size_type x=0;x<width;x++){
      for(size_type y=0;y<(height+1)/2;y++){
	double r1,g1,b1,r2,g2,b2;
	Image_container_traits::get_value(*target,x,height-1-y,&r1,&g1,&b1);
	Image_container_traits::get_value(*target,x,y,&r2,&g2,&b2);
	Image_container_traits::set_value(target,x,y,r1,g1,b1);
	Image_container_traits::set_value(target,x,height-1-y,r2,g2,b2);
      }
    }
  }

  
/*

  ##########################
  # image_container_traits #
  ##########################

 */

  //! Namespace that must contain the functions get_value() ! and
  //set_value() adapted to the used data structure. See the example on
  //the main page.
  namespace Image_container_traits{    
  } //END OF namespace Image_container_traits
 
} //END OF namespace Image_file

#endif



